# -*- coding: utf-8 -*-
# copyright: (c) 2020 by Jesse Johnson.
# license: MPL-2.0, see LICENSE for more details.
'''Provide CLI management.'''

from argufy import Parser

from proman_versioning.cli import version, release


def main() -> None:
    '''Provide CLI for git-tools.'''
    parser = Parser(command_type='subcommand')
    parser.add_commands(release)
    parser.add_commands(version)
    parser.dispatch()


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# copyright: (c) 2020 by Jesse Johnson.
# license: MPL-2.0, see LICENSE for more details.
"""Provide configuration for versioning tools."""

import os
from dataclasses import dataclass, field

# from pprint import pprint
from typing import Optional, Tuple
from urllib.parse import urljoin, urlparse

from compendium.loader import ConfigFile
from git.config import GitConfigParser

INDEX_URL = urlparse('https://pypi.org')
VENV_PATH = os.getenv('VIRTUAL_ENV', None)
PATHS = [VENV_PATH] if VENV_PATH else []

# TODO check VCS for paths
basedir = os.getcwd()
filenames = ['pyproject.toml', 'setup.cfg']
# pyproject_path = os.path.join(basedir, 'pyproject.toml')
# lock_path = os.path.join(basedir, 'proman-lock.json')
pypackages_dir = os.path.join(basedir, '__pypackages__')

grammar: str = os.path.join(
    os.path.dirname(__file__), 'grammars', 'conventional_commits.lark'
)
# 'proman_versioning/templates/gitmessage.j2'


@dataclass
class TaskRunner:
    """Manager task runner setup."""

    name: str = 'invoke'
    template: str = field(init=False)
    templates_dir: str = os.path.join(os.path.dirname(__file__), 'templates')

    def __post_init__(self) -> None:
        """Run post initialization."""
        self.template = f"{self.name}_hooks.j2"


@dataclass
class HooksConfig(TaskRunner):
    """Manage hooks config."""

    hooks_dir: str = os.path.join(basedir, '.git', 'hooks')
    hooks: Tuple[str, ...] = (
        'applypatch-msg',
        'pre-applypatch',
        'post-applypatch',
        'pre-commit',
        'pre-merge-commit',
        'prepare-commit-msg',
        'post-commit',
        'pre-rebase',
        'post-checkout',
        'post-merge',
        'pre-push',
        'pre-recieve',
        'update',
        'post-recieve',
        'post-update',
        'reference-transaction',
        'push-to-checkout',
        'pre-auto-gc',
        'post-rewrite',
        'rebase',
        'sendemail-validate',
        'fsmonitor-watchman',
        'p4-changelist',
        'p4-prepare-changelist',
        'p4-post-changelist',
        'p4-pre-submit',
        'post-index-change',
    )


@dataclass
class GitConfig:
    """Manage git config."""

    system_config: str = os.path.join(os.sep, 'etc', 'gitconfig')
    global_config: str = os.path.join(os.path.expanduser('~'), '.gitconfig')

    def __post_init__(self) -> None:
        """Run post initiation."""
        self.load()

    def load(self) -> None:
        """Load git configuration."""
        with GitConfigParser(self.global_config, read_only=True) as cfg:
            cfg.read()
            if not cfg.has_section('commit'):
                cfg.add_section('commit')
            # pprint(cfg.__dict__)  # ._sections)

    def save(self) -> None:
        """Save git configuration."""
        with GitConfigParser(self.global_config, read_only=False) as cfg:
            if not cfg.has_section('commit'):
                cfg.add_section('commit')
            cfg.write()


@dataclass
class Config(ConfigFile):
    """Manage settings from configuration file."""

    filepath: str
    index_url: str = urljoin(INDEX_URL.geturl(), 'simple')
    python_versions: tuple = ()
    digest_algorithm: str = 'sha256'
    include_prereleases: bool = False
    lookup_memory: Optional[str] = None
    writable: bool = True

    def __post_init__(self) -> None:
        """Initialize settings from configuration."""
        super().__init__(self.filepath)
        if os.path.isfile(self.filepath):
            self.load()



# -*- coding: utf-8 -*-
# copyright: (c) 2021 by Jesse Johnson.
# license: MPL-2.0, see LICENSE for more details.
"""Convenience tools to manage Git projects with Python."""

import logging
import os
from typing import Any, List, Union

from pygit2 import Repository

from proman_versioning import config, exception

# from proman_versioning.config import Config
from proman_versioning.controller import IntegrationController
from proman_versioning.grammars.conventional_commits import CommitMessageParser
from proman_versioning.vcs import Git
from proman_versioning.version import PythonVersion

logging.getLogger(__name__).addHandler(logging.NullHandler())

__author__ = 'Jesse P. Johnson'
__author_email__ = 'jpj6652@gmail.com'
__title__ = 'proman-releases'
__description__ = 'Convenience module to manage VCS tools with Python.'
__version__ = '0.1.1-alpha.1'
__license__ = 'MPL-2.0'
__copyright__ = 'Copyright 2021 Jesse Johnson.'


def get_repo(path: str = os.getcwd()) -> Git:
    """Load the repository object."""
    return Git(Repository(os.path.join(path)))


def get_source_tree(
    basepath: str = os.getcwd(), filenames: List[str] = config.filenames
) -> config.Config:
    """Get source tree from path."""
    for filename in filenames:
        filepath = os.path.join(basepath, filename)
        if os.path.isfile(filepath):
            return config.Config(filepath=filepath)
    raise exception.PromanWorkflowException('no configuration found')


def get_python_version(cfg: Union[config.Config, str]) -> PythonVersion:
    """Get python version from configurations."""
    if isinstance(cfg, config.Config):
        version = (
            cfg.retrieve('/tool/proman/release/version')
            or cfg.retrieve('/tool/proman/version')
            or cfg.retrieve('/tool/poetry/version')
            or cfg.retrieve('/metadata/version')
        )
        if version is None:
            raise exception.PromanWorkflowException('no version found')

        version = PythonVersion(
            version=version,
            enable_devreleases=cfg.retrieve(
                '/tool/proman/release/enable_devreleases', False
            ),
            enable_prereleases=cfg.retrieve(
                '/tool/proman/release/enable_prereleases', False
            ),
            enable_postreleases=cfg.retrieve(
                '/tool/proman/release/enable_postreleases', False
            ),
        )
    else:
        version = PythonVersion(cfg)
    return version


def get_release_controller(*args: Any, **kwargs: Any) -> IntegrationController:
    """Create and return a release controller."""
    basepath = kwargs.get('basepath', os.getcwd())
    filenames = kwargs.get('filenames', config.filenames)
    cfg = get_source_tree(basepath=basepath, filenames=filenames)
    version = get_python_version(kwargs.pop('version', cfg))
    return IntegrationController(
        version=version,
        config=cfg,
        repo=get_repo(basepath),
        **kwargs,
    )


repo = get_repo()
source_tree = get_source_tree()
parser = CommitMessageParser()

__all__ = [
    'get_repo',
    'get_source_tree',
    'get_release_controller',
    'parser',
    'repo',
    'source_tree',
]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['proman_versioning', 'proman_versioning.cli', 'proman_versioning.grammars']

package_data = \
{'': ['*'], 'proman_versioning': ['templates/*']}

install_requires = \
['argufy>=0.1.1-alpha.12,<0.2.0',
 'compendium>=0.1.1-alpha.2,<0.2.0',
 'jinja2>=2.11.2,<3.0.0',
 'lark-parser>=0.10.0,<0.11.0',
 'packaging>=20.9,<21.0',
 'pygit2>=1.6.1,<2.0.0',
 'transitions>=0.8.4,<0.9.0']

entry_points = \
{'console_scripts': ['proman-version = proman_versioning.__main__:main']}

setup_kwargs = {
    'name': 'proman-versioning',
    'version': '0.1.1a1',
    'description': 'Project Manager Versioning tool.',
    'long_description': '# Proman Versioning\n\n[![License](https://img.shields.io/badge/License-Apache%202.0-blue.svg)](https://spdx.org/licenses/MPL-2.0)\n[![Build Status](https://travis-ci.org/kuwv/proman-versioning.svg?branch=master)](https://travis-ci.org/kuwv/proman-versioning)\n[![codecov](https://codecov.io/gh/kuwv/proman-versioning/branch/master/graph/badge.svg)](https://codecov.io/gh/kuwv/proman-versioning)\n\n## Overview\n\nProject Manager Versioning tool.\n\n## Install\n\n`pip install proman-versioning`\n\n## Setup\n\n```\n[tool.proman.versioning]\nenable_devreleases = true\nenable_prereleases = true\nenable_postreleases = true\n\n[[tool.proman.versioning.files]]\nfilepath = "example/__init__.py"\npattern = "__version__ = \'${version}\'"\n\n[[tool.proman.versioning.files]]\nfilepath = "pyproject.toml"\npattern = "version = \\"${version}\\""\n```\n',
    'author': 'Jesse P. Johnson',
    'author_email': 'jpj6652@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.2,<4.0.0',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-
# copyright: (c) 2020 by Jesse Johnson.
# license: MPL-2.0, see LICENSE for more details.
'''Parse Git commit messages.'''

# import logging
import os
from typing import Any, List, Optional

from pygit2 import GIT_OBJ_COMMIT, Commit, Repository, Signature, Tag

# from proman_versioning import exception

# from transitions import Machine


class Git:
    '''Provide settings for git repositories.'''

    system_config: str = os.path.join(os.sep, 'etc', 'gitconfig')
    global_config: str = os.path.join(os.path.expanduser('~'), '.gitconfig')

    def __init__(self, repo: Repository) -> None:
        '''Initialize git object.'''
        self.repo = repo
        self.ref = 'HEAD'
        self.hooks_dir = os.path.join(self.repo.path, 'hooks')
        self.config = os.path.join(self.repo.path, 'config')

    @property
    def branch(self) -> str:
        '''Retrieve the current branch.'''
        return self.repo.head.name

    def commit(
        self,
        name: Optional[str] = None,
        filepaths: List[str] = [],
        **kwargs: Any,
    ) -> Commit:
        '''Create commit.'''
        if not name:
            name = f"refs/heads/{self.ref}"
        author = Signature('Jesse P. Johson', 'jpj6652@gmail.com')
        committer = kwargs.get('commiter', author)
        message: str = kwargs.get('message', 'ci: generated commit')

        # populate index
        index = self.repo.index
        if filepaths == []:
            index.add_all()
        else:
            for filepath in filepaths:
                index.add(
                    os.path.relpath(
                        filepath, os.path.join(self.repo.path, '..')
                    )
                )
        index.write()
        tree = index.write_tree()

        # get parent
        parents = kwargs.get('parents', [self.repo.head.target])
        encoding = kwargs.get('encoding', 'utf-8')

        # commit
        if not kwargs.get('dry_run', False):
            commit = self.repo.create_commit(
                name,
                author,
                committer,
                message,
                tree,
                parents,
                encoding,
            )
        return commit

    def tag(self, name: str, ref: str = 'HEAD', **kwargs: Any) -> Tag:
        '''Create tag.'''
        commit = self.repo.resolve_refish(ref)[0]
        oid = commit.hex
        kind = kwargs.get('kind', GIT_OBJ_COMMIT)
        signature = kwargs.get('signature', commit.signature)
        message = kwargs.get('message', f"ci: {name}")
        if not kwargs.get('dry_run', False):
            tag = self.repo.create_tag(name, oid, kind, signature, message)
        return tag



# -*- coding: utf-8 -*-
# copyright: (c) 2020 by Jesse Johnson.
# license: MPL-2.0, see LICENSE for more details.
"""Parse Git commit messages."""

# import logging
import os
import re
import sys
from copy import deepcopy
from string import Template
from typing import Any

from proman_versioning import exception
from proman_versioning.config import Config
from proman_versioning.grammars.conventional_commits import CommitMessageParser
from proman_versioning.vcs import Git
from proman_versioning.version import PythonVersion

# from packaging.version import Version
# from transitions import Machine

# TODO: version comparison against previous version
# has API spec been modified?
# has Python version changed?
# has requirements versions changed?


# TODO determine relation with state and git hooks
class IntegrationController(CommitMessageParser):
    """Control version releases."""

    # kinds = ['rolling', 'sustainment']
    # Trunk Based Development (TBD) - GitHub flow
    # Stage Based Development (SBD) - use of DTAP
    # Release Branching Strategy (RBS) - PEP440
    # Feature Branching Strategy (FBS) - Agile

    def __init__(
        self,
        version: PythonVersion,
        config: Config,
        repo: Git,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Initialize commit message action object."""
        self.version = version
        self.config = config
        parse_current_repo = kwargs.pop('parse_current_repo', True)
        super().__init__(*args, **kwargs)

        self.vcs = repo
        if parse_current_repo:
            head = self.vcs.repo.head
            target = self.vcs.repo[head.target]
            self.parse(target.message)

    @staticmethod
    def __update_config(
        filepath: str,
        version: str,
        new_version: str,
        dry_run: bool = False,
    ) -> None:
        """Update config file with new file."""
        # TODO: if file does not exist
        with open(filepath, 'r+') as f:
            file_contents = f.read()
            pattern = re.compile(re.escape(version), flags=0)
            file_contents = pattern.sub(new_version, file_contents)
            if not dry_run:
                try:
                    f.seek(0)
                    f.truncate()
                    f.write(file_contents)
                except Exception as err:
                    print(err, file=sys.stderr)
            else:
                print(file_contents, file=sys.stdout)

    def update_configs(
        self, new_version: PythonVersion, **kwargs: Any
    ) -> None:
        """Update version within config files."""
        stats = self.vcs.repo.diff('HEAD').stats
        if stats.files_changed == 0:
            if str(self.version) != str(new_version):
                dry_run = kwargs.pop('dry_run', False)
                filepaths = (
                    self.config['tool']['proman']['versioning']['files']
                )
                for filepath in filepaths:
                    self.__update_config(
                        filepath=os.path.abspath(
                            os.path.join(
                                self.vcs.repo.path,
                                '..',
                                filepath['filepath']
                            )
                        ),
                        version=(
                            Template(filepath['pattern']).substitute(
                                version=str(self.version)
                            )
                        ),
                        new_version=(
                            Template(filepath['pattern']).substitute(
                                version=new_version
                            )
                        ),
                        dry_run=dry_run,
                    )
                self.version = new_version
                if kwargs.pop('commit', True):
                    self.vcs.commit(
                        filepaths=[f['filepath'] for f in filepaths],
                        message=(f"ci(version): apply {new_version} updates"),
                        dry_run=dry_run,
                    )
                if kwargs.get('tag', False):
                    self.vcs.tag(
                        name=str(new_version),
                        ref='HEAD',
                        message=None,
                        dry_run=dry_run,
                    )
            else:
                raise exception.PromanWorkflowException(
                    'no new version available'
                )
        else:
            raise exception.PromanWorkflowException(
                'git repository is not clean'
            )

    def start_release(self, kind: str = 'dev', **kwargs: Any) -> str:
        """Start a release."""
        new_version = deepcopy(self.version)
        if kind == 'dev':
            new_version.start_devrelease()  # type: ignore
        elif kind == 'pre':
            new_version.start_prerelease()  # type: ignore
        self.update_configs(new_version, **kwargs)
        return str(new_version)

    def finish_release(self, **kwargs: Any) -> str:
        """Start a release."""
        new_version = deepcopy(self.version)
        new_version.finish_release()  # type: ignore
        self.update_configs(new_version, **kwargs)
        return str(new_version)

    @staticmethod
    def __bump_release(version: PythonVersion) -> PythonVersion:
        """Update release number."""
        if version.is_devrelease:
            version.bump_devrelease()
        elif version.is_prerelease:
            version.bump_prerelease()
        elif version.is_postrelease:
            version.bump_postrelease()
        elif version.enable_postreleases:
            version.start_postrelease()  # type: ignore
        return version

    def bump_version(self, **kwargs: Any) -> str:
        """Update the version of the application."""
        new_version = deepcopy(self.version)

        # local number depends on metadata / fork / conflict existing vers
        if self.title['break'] or self.footer['breaking_change']:
            new_version.bump_major()
        elif 'type' in self.title:
            if self.title['type'] == 'feat':
                new_version.bump_minor()
            elif self.title['type'] == 'fix':
                new_version.bump_micro()

            # update release instance
            elif self.title['type'] == 'build':
                new_version = self.__bump_release(new_version)
            elif self.title['type'] == 'ci':
                new_version = self.__bump_release(new_version)
            elif self.title['type'] == 'docs':
                new_version = self.__bump_release(new_version)
            elif self.title['type'] == 'perf':
                new_version = self.__bump_release(new_version)
            elif self.title['type'] == 'refactor':
                new_version = self.__bump_release(new_version)
            elif self.title['type'] == 'style':
                new_version = self.__bump_release(new_version)
            elif self.title['type'] == 'test':
                new_version = self.__bump_release(new_version)
            elif self.title['type'] == 'chore':
                new_version = self.__bump_release(new_version)

        self.update_configs(new_version, **kwargs)
        return str(new_version)



# -*- coding: utf-8 -*-
# copyright: (c) 2020 by Jesse Johnson.
# license: MPL-2.0, see LICENSE for more details.
'''Provide exceptions.'''


class PromanWorkflowException(Exception):
    '''Provide base errors in PackageManager.'''
